/****************************************************************************
*
*  Motor.c
*
*  Motor Control Code
*
****************************************************************************/

#pragma chip PIC16F877

// ---- Include Files -------------------------------------------------------

#include "Motor.h"
#include "Delay.h"
#include "Puts.h"
#include "Sensor.h"

// ---- Public Variables ----------------------------------------------------

#define HITEC_SERVOS    0

uns8  gSpeedL;
uns8  gSpeedR;
uns8  gPrevSpeedL;
uns8  gPrevSpeedR;

#define DO_RAMP         0

uns8    gRampDelta;

#if DO_RAMP

#define RAMP_DELTA  gRampDelta

uns8  gActualSpeedFixedL;
uns8  gActualSpeedFixedR;

#endif  // DO_RAMP

// ---- Private Constants and Types -----------------------------------------
// ---- Private Variables ---------------------------------------------------

#pragma  bit SERVO_R    @ PORTB.1
#pragma  bit SERVO_L    @ PORTB.2

// ---- Private Function Prototypes -----------------------------------------
// ---- Functions -----------------------------------------------------------

#define SPEED_OFF_FIXED ( SPEED_OFF << RAMP_SHIFT )

#if DO_RAMP
uns8 AdjustSpeed( uns8 desiredSpeedFixed, uns8 actualSpeedFixed )
{
    if (( desiredSpeedFixed < actualSpeedFixed ) && ( desiredSpeedFixed >= SPEED_OFF_FIXED ))
    {
        // The speed is being reduced

        return desiredSpeedFixed;
    }

    if (( desiredSpeedFixed > actualSpeedFixed ) && ( desiredSpeedFixed <= SPEED_OFF_FIXED ))
    {
        // The speed is being reduced

        return desiredSpeedFixed;
    }

    if ( desiredSpeedFixed > actualSpeedFixed )
    {
        actualSpeedFixed += RAMP_DELTA;
    }
    else
    if ( desiredSpeedFixed < actualSpeedFixed )
    {
        actualSpeedFixed -= RAMP_DELTA;
    }

    return actualSpeedFixed;
}
#endif

extern void InitMotors( void )
{
    gPrevSpeedL = 0;
    gPrevSpeedR = 0;
    gRampDelta  = 2;

#if DO_RAMP
    gActualSpeedFixedL = 15 << RAMP_SHIFT;
    gActualSpeedFixedR = 15 << RAMP_SHIFT;
#endif
}

/****************************************************************************
*
*  Sets the motor speed of both motors.
*/

extern void PulseMotors( void )
{
    uns8    pulseL;
    uns8    pulseR;

#if DO_RAMP

    uns8    gActualSpeedL;
    uns8    gActualSpeedR;

    uns8    gDesiredSpeedFixedL = gSpeedL << RAMP_SHIFT;
    uns8    gDesiredSpeedFixedR = gSpeedR << RAMP_SHIFT;

    gActualSpeedFixedL = AdjustSpeed( gDesiredSpeedFixedL,  gActualSpeedFixedL );
    gActualSpeedFixedR = AdjustSpeed( gDesiredSpeedFixedR,  gActualSpeedFixedR );

    gActualSpeedL = ( gActualSpeedFixedL >> RAMP_SHIFT );
    gActualSpeedR = ( gActualSpeedFixedR >> RAMP_SHIFT );

#else

#define gActualSpeedL   gSpeedL
#define gActualSpeedR   gSpeedR

#endif

    // We need to run the right motor in reverse, so we
    // do the adjustment here.
    //
    // We're converting a number in the range 10-20 into 20-10
    // so subtracting from 30 does the right thing.

#if HITEC_SERVOS

    // Hitec servos seem to run in the opposite direction from the GWS servos

    pulseL = 300 - gActualSpeedL;
    pulseR = gActualSpeedR;
#else
    pulseL = gActualSpeedL;
    pulseR = 300 - gActualSpeedR;
#endif
    
    //puts( "PulseMotors called\n" );

   // We send both pulses at the same time, so we need to
   // figure out which one is shorter.

   if ( pulseL <= pulseR )
   {
      gT0 = pulseL;
      gT1 = pulseR - pulseL;
   }
   else
   {
      gT0 = pulseR;
      gT1 = pulseL - pulseR;
   }

#if DEBUG
   if (( gSpeedL != gPrevSpeedL ) || ( gSpeedR != gPrevSpeedR ))
   {
      puts( "PulseMotors L:" );
      putu8( gSpeedL );
      puts( " R:" );
      putu8( gSpeedR );
      puts( " t0:" );
      putu8( gT0 );
      puts( " t1:" );
      putu8( gT1 );
      puts( " C: " );
      putu16( gHuntCounter );
      puts( "\n" );

      gPrevSpeedL = gSpeedL;
      gPrevSpeedR = gSpeedR;
   }
#endif

   // Start both pulses
   //
   // Because the servos have some capacitance, we need to avoid back-to-back
   // read-modify-write cycles to the same port. Since the servos are bit
   // fields, setting either one is inherently a read-modify-write cycle.
   // The nop introduces enough delay to allow the 2nd read to get the
   // correct value.

   //puts( "Starting delay\n" );

   SERVO_L = 1;  nop();
   SERVO_R = 1;

   Delay10uSec( gT0 );

   //puts( "Finished delay\n" );

   if ( gT1 != 0 )
   {
      if ( pulseL < pulseR )
      {
         // Left pulse stops first.

         SERVO_L = 0;
      }
      else
      {
         // Right pulse stops first.

         SERVO_R = 0;
      }

      Delay10uSec( gT1 );
   }

   //puts( "Finished second delay\n" );

   SERVO_L = 0; nop();
   SERVO_R = 0;

} // PulseMotors

