/****************************************************************************
*
*  ServoTst.c  
*
*  Test code for driving the servos
*
****************************************************************************/

#pragma chip PIC16F877

// The BootLoader requires a GOTO at address 0x0003 rather than 0x0000

#pragma resetVector 0x0003

#pragma  bit SERVER_R @ PORTB.1
#pragma  bit SERVER_L @ PORTB.2

/***************************************************************************/
/**
*  Writes a single character to the serial port.
*/

void putc( char ch )
{
   // Wait for space in the TX buffer

   while ( TXIF == 0 ) ;

   TXREG = ch;

} // putc

/***************************************************************************/
/**
*  Writes a string the serial port.
*/

void puts( const char *str )
{
   while ( *str )
   {
      // Wait for space in the TX buffer

      while ( TXIF == 0 ) ;

      TXREG = *str++;
   }

} // puts

/***************************************************************************/
/**
*  Delays for a multiple of 100 microseconds.
*
*  @note This routine assumes a 20 MHz clock.
*/

void Delay100uSec
(
   uns8 num100uSec   /**< Number of 100 microsecond units to delay         */
)
{
   uns8 inner;

#asm
OLOOP    movlw    D'124'         ;load inner loop counter constant
         movwf    inner          ;load counter

ILOOP    clrwdt                  ;kick the dog
         decfsz   inner,F        ;has inner loop expired?
         goto     ILOOP          ;no, repeat inner loop

         decfsz   num100uSec,F   ;yes, has outer loop expired?
         goto     OLOOP          ;no, repeat outer loop
#endasm

} // Delay100uSec

/***************************************************************************/
/**
*  Sets the motor speed of both motors.
*/

void PulseMotors( uns8 speedL, uns8 speedR )
{
   uns8  t0;
   uns8  t1;

#if DEBUG
   puts( "PulseMotors L:" );
   putu8( speedL );
   puts( " R:" );
   putu8( speedR );
   puts( "\r\n" );
#endif

   if ( speedL <= speedR )
   {
      t0 = speedL;
      t1 = speedR - speedL;
   }
   else
   {
      t0 = speedR;
      t1 = speedL - speedR;
   }

   // Start both pulses
   //
   // Because the servos have some capacitance, we need to avoid back-to-back
   // read-modify-write cycles to the same port. Since the servos are bit
   // fields, setting either one is inherently a read-modify-write cycle.
   // The nop introduces enough delay to allow the 2nd read to get the
   // correct value.


   SERVER_L = 1;  nop();
   SERVER_R = 1;

   Delay100uSec( t0 );

   if ( t1 != 0 )
   {
      if ( speedL < speedR )
      {
         SERVER_L = 0;
      }
      else
      {
         SERVER_R = 0;
      }

      Delay100uSec( t1 );
   }

   SERVER_L = 0; nop();
   SERVER_R = 0;

} // PulseMotors

/***************************************************************************/
/**
*  Wait for Timer 0 to rollover
*/

void WaitForTimer0Rollover( void )
{
#asm
_WaitForTimer0Rollover:
	clrwdt
	btfss	INTCON,T0IF             // Has rollover occured?
	goto	_WaitForTimer0Rollover  // No - keep looping

	bcf		INTCON,T0IF          // Reset Timer
#endasm

} // WaitForTimer0Rollover

/***************************************************************************/
/**
*  Main program for the Mini Sumo
*/

void main( void )
{
   uns8  speedL;
   uns8  speedR;
   uns8  ch;

   // RB0   I /INT
   // RB1   O PWM1/ServoR
   // RB2   O PWM2/ServoL
   // RB3   O PWM3
   // RB4   O PWM4
   // RB5   I B5
   // RB6   I XAxis/CTS
   // RB7   I YAxis/RTS

   PORTB = 0b00000000;
   TRISB = 0b11100001;

   // OPTION register
   //
   // 7  /RBPU    Port B Pull-up enable (1 = disabled, 0 = enabled)
   // 6  INTEDG   Interrupt Edge Select, 1 = rising edge of RB0/INT pin
   // 5  T0CS     TMR0 Clock Select, 0 = Fosc/4, 1 = RA4/T0CKI pin
   // 4  T0SE     TMR0 Source Edge, 0 = rising 1 = falling (on RA4/T0CKI pin)
   // 3  PSA      Prescalar assignment, 0 = PS for TMR0, 1 = PS for WatchDogTimer
   // 2  PS2      0-7 chooses 2 thru 256 for TMR0, 1 thru 128 for WDT
   // 1  PS1
   // 0  PS0

   OPTION = 0b01000111; // TMR0 = 76.3 ticks per second

   // We assume that the Boot Loader has already initialized the serial port
   // so we don't need to touch the baud rate or anything.

   // The boot loader leaves the cursor after the words "Jumping to user code"
   // but on the next line. Send a CR to reposition to the beginning of the
   // line.

   puts( "\r" );

   // Output our test string

   puts( "Hello World\r\n" );

   speedL = 15;
   speedR = 15;

   while ( 1 )
   {
      WaitForTimer0Rollover();

      if ( RCIF == 1 )
      {
         ch = RCREG;
         if (( ch >= '0' ) && ( ch <= '9' ))
         {
            putc( ch );
            
            ch -= '0';
            speedL = 10;
            speedL += ch;
            speedR = speedL;
         }
      }
      PulseMotors( speedL, speedR );   
   }

   // The serial transmitter has a TX buffer of length 1, and a shift
   // register. We need to wait until the TX buffer is empty and one more
   // character time to ensure that the last character that we sent 
   // actually clears the shift register before we go to sleep.
   // The simplest way to do that is to send two dummy characters. The first
   // character will fill the shift register, and the 2nd will fill the TX
   // buffer. Since the PIC goes to sleep immediately on returning from main
   // neither of these characters should actually clear the chip.

   puts( "xx" );
}

